import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import Node, is_consistent, is_value_at_index


def test__insert_before_node_one_in_list_insert_not_node_failure():
    # Setup
    node0: str = "string"
    cdll0: CDLL = CDLL(values=[None])

    # Validation
    with pytest.raises(AttributeError):
        # noinspection PyTypeChecker
        cdll0._insert_before_node(anchor=cdll0._head, insert=node0)


def test__insert_before_node_one_in_list_insert_one_and_reassign_head_success():
    # Setup
    data0: int = 0
    data1: int = 1
    values0: list[int] = [data0]
    values1: list[int] = [data1, data0]
    node0: Node = Node(value=data1)
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0._insert_before_node(anchor=cdll0._head, insert=node0)

    # Validation
    assert cdll0._last.next is node0
    assert cdll0._last.previous is node0
    assert node0.next is cdll0._last
    assert node0.previous is cdll0._last
    assert len(cdll0) == 2
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test__insert_before_node_two_in_list_insert_one_in_middle_success():
    # Setup
    data0: int = 0
    data1: int = 1
    data2: int = 2
    values0: list[int] = [data0, data1]
    values1: list[int] = [data0, data2, data1]
    node0: Node = Node(value=data2)
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0._insert_before_node(anchor=cdll0._last, insert=node0)

    # Validation
    assert cdll0._head.next is node0
    assert cdll0._last.previous is node0
    assert node0.next is cdll0._last
    assert node0.previous is cdll0._head
    assert len(cdll0) == 3
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test__insert_before_node_four_in_list_insert_one_in_middle_success():
    # Setup
    data0: int = 0
    data1: int = 1
    data2: int = 2
    data3: int = 3
    data4: int = 4
    values0: list[int] = [data0, data1, data3, data4]
    values1: list[int] = [data0, data1, data2, data3, data4]
    node0: Node = Node(value=data2)
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0._insert_before_node(anchor=cdll0._last.previous, insert=node0)

    # Validation
    assert cdll0._head.next.next is node0
    assert cdll0._last.previous.previous is node0
    assert node0.next.next is cdll0._last
    assert node0.previous.previous is cdll0._head
    assert len(cdll0) == 5
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)



import pytest

from src.py_cdll import CDLL, ValueNotFoundError, MultipleValuesFoundError
from src.py_cdll.node import is_consistent, is_value_at_index


def test_replace_unique_present_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    values0: list[str] = [data0, data1, data2]
    values1: list[str] = [data0, data, data2]
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0.replace_unique(value=data1, replacement=data)

    # Validation
    assert cdll0[1] is data
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test_replace_unique_not_present_failure():
    # Setup
    data_find: str = "data_find"
    data_new: str = "data_new"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    values0: list[str] = [data0, data1, data2]
    cdll0: CDLL = CDLL(values=values0)

    # Validation
    with pytest.raises(ValueNotFoundError):
        cdll0.replace_unique(value=data_find, replacement=data_new)


def test_replace_unique_multiple_options_failure():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    values0: list[str] = [data0, data1, data1, data2]
    cdll0: CDLL = CDLL(values=values0)

    # Validation
    with pytest.raises(MultipleValuesFoundError):
        cdll0.replace_unique(value=data1, replacement=data)



from typing import TypeVar, Iterator, Sequence, Callable

from .exceptions import NotANodeError, NodesNotInSameListError, NoNewHeadError, IndexNotFoundError, IndexOutOfRangeError

_T = TypeVar("_T")


class Node:
    """Circular Doubly Linked Node"""

    def __init__(self, value: _T) -> None:
        self.head: Node = self
        self.next: Node = self
        self.previous: Node = self
        self.index: int = 0
        self.value: _T = value

    def __repr__(self) -> str:
        values: list[str] = [repr(node.value) for node in nodes(head=self)]
        string = f"Node(...{id(self) % 1000000})[{length(node=self)}, {[self.value]}{values[1:]}]"
        return string


def nodes_from_values(values: Sequence[_T]) -> Node | None:
    """Create circular doubly linked list of Nodes containing values."""
    values_iterator: Iterator[_T] = iter(values)

    try:
        head: Node | None = Node(value=next(values_iterator))
        current: Node | None = head
    except StopIteration:
        head, current = None, None

    for value in values_iterator:
        node: Node = Node(value=value)
        insert_between(before=current, after=head, insert=node)
        current = current.next

    return head


def nodes(head: Node, callback: Callable[[], Node] | None = None) -> Iterator[Node]:
    current: Node = head
    steps: int = 0

    is_callback_callable: bool = isinstance(callback, Callable)

    while True:
        yield current

        current = current.next
        steps += 1

        is_node_in_list: bool = current.head is callback() if is_callback_callable else current.head is head.head

        if not is_node_in_list:
            if callback:
                head = callback()
                new_current: Node = head
            else:
                new_current: Node = head.head

            for _ in range(steps):
                new_current = new_current.next

            if new_current.index != steps:
                break

            current = new_current

        if current is head:
            break


def is_consistent(node: Node) -> bool:
    is_correct: bool = True
    head: Node = node.head
    nodes__: list[Node] = []

    current: Node = head
    while True:
        nodes__.append(current)
        current = current.next
        if current is head:
            break

    for index, node in enumerate(nodes__):
        is_correct &= node.head is head
        is_correct &= node.index == index
        is_correct &= node.next.index == (node.index + 1) % len(nodes__)
        is_correct &= node.previous.index == (node.index - 1) % len(nodes__)
        is_correct &= node.next is nodes__[(index + 1) % len(nodes__)]
        is_correct &= node.previous is nodes__[(index - 1) % len(nodes__)]
        is_correct &= node.next.previous is node
        is_correct &= node.previous.next is node

        if not is_correct:
            break

    return is_correct


def node_with_index(node: Node, index: int) -> Node:
    if index >= length(node):
        raise IndexOutOfRangeError(f"Index '{index}' is out of range '{length(node) - 1}'.")

    head: Node = node.head
    current: Node = head

    while True:
        if current.index == index:
            break

        current = current.next

        if current is head:
            raise IndexNotFoundError(f"Index '{index}' not found in Node sequence.")

    return current


def is_value_at_index(node: Node, reference: list) -> bool:
    is_correct: bool = True

    if length(node=node) != len(reference):
        is_correct = False

    for index, value in enumerate(reference):
        try:
            node_at_index: Node = node_with_index(node=node, index=index)
            is_correct &= node_at_index.value == value
        except (IndexOutOfRangeError, IndexNotFoundError):
            is_correct = False

        if not is_correct:
            break

    return is_correct


def length_step(node: Node):
    """
    Count nodes in circular doubly linked sequence by stepping through a full round.
    Necessary in contexts where Node indices are not guaranteed to be correct.
    """

    current: Node = node
    amount: int = 0

    while True:
        amount += 1
        current = current.next
        if current is node:
            break

    return amount


def length(node: Node) -> int:
    """
    Calculates Node amount in circular doubly linked list by index of last Node.
    Assumes Node indices are correct.
    """

    try:
        length_: int = node.head.previous.index + 1
    except AttributeError as exception:
        raise NotANodeError(f"'{node}' is not a Node, but a '{type(node)}'.") from exception

    return length_


def remove(target: Node) -> Node:
    """
    Requires a minimum length of 2 nodes
    Assumes that target node is in a fully circular doubly linked list.
    """

    head: Node | None = None

    if target is target.head:
        head = target.next

    removed: Node = insert_between(before=target.previous, after=target.next, insert=None, head=head)

    return removed


def equal(first: Node, second: Node) -> bool:
    """Compare nodes from head in two circular doubly linked sequences."""
    equality: bool = True

    first_head: Node = first
    second_head: Node = second

    first_current: Node = first_head
    second_current: Node = second_head

    while True:
        equality &= first_current.value is second_current.value or first_current.value == second_current.value
        equality &= first_current.index == second_current.index
        equality &= first_current.head is first_head and second_current.head is second_head

        if not equality:
            break
        elif first_current.next is first_head and second_current.next is second_head:
            break
        elif first_current.next is first_head and second_current.next is not second_head:
            equality &= False
            break
        elif first_current.next is not first_head and second_current.next is second_head:
            equality &= False
            break

        first_current = first_current.next
        second_current = second_current.next

    return equality


def reverse_order(head: Node) -> Node:
    reversed_sequence: Node = head

    if head.next is not head:
        reversed_sequence: Node = head.previous
        previous_node: Node = head
        current_node: Node = head.previous

        while True:
            old_previous: Node = current_node.previous

            current_node.next = current_node.previous
            current_node.previous = previous_node

            previous_node = current_node
            current_node = old_previous

            if previous_node is head:
                break

    update_head(node=reversed_sequence)

    return reversed_sequence


def split_head_from_tail(node: Node) -> tuple[Node, Node | None]:
    first_head: Node = node

    if node.next is node:
        second_head: None = None
    else:
        second_head: Node = node.next
        split(first_head=first_head, second_head=second_head)

    return first_head, second_head


def middle_adjacent(head: Node) -> tuple[Node, Node]:
    """When node amount is uneven, preferentially adds one more node to before than after."""
    slow: Node = head
    fast: Node = head

    while fast.next is not head and fast.next.next is not head:
        slow = slow.next
        fast = fast.next.next

    before_last: Node = slow
    after_head: Node = slow.next

    return before_last, after_head


def stitch(head: Node, last: Node) -> None:
    """Stitch together head and last to make sequence circular."""
    head.previous, last.next = last, head


def split(first_head: Node, second_head: Node) -> None:
    """Split one circular doubly linked list into two."""

    if first_head.head is not second_head.head:
        raise NodesNotInSameListError(f"Nodes have different heads and are thus in different lists.")

    first_last: Node = second_head.previous
    second_last: Node = first_head.previous

    stitch(head=first_head, last=first_last), stitch(head=second_head, last=second_last)

    # TODO: refactor to extract repetitive code
    current: Node = first_head
    index: int = 0
    while True:
        current.head = first_head
        current.index = index

        current = current.next
        index += 1

        if current is first_head:
            break

    current: Node = second_head
    index: int = 0
    while True:
        current.head = second_head
        current.index = index

        current = current.next
        index += 1

        if current is second_head:
            break


def insert_between(before: Node, after: Node, insert: Node | None, head: Node | None = None) -> Node | None:
    """
    Insert sequence between before and after anchors. Return sequence previously between before and after anchors.
    Requires a minimum length of 1 Node.
    When working with 1 Node it is always going to assume that it should operate between its connections.
    Allows that insert can have multiple connected Nodes.
    Assumes that each input Node sequence is in a fully circular doubly linked list.
    """

    is_single_node_remaining: bool = before.index == after.index
    is_remaining_node_head: bool = before.index == 0
    is_new_head_provided: bool = head is not None

    if is_single_node_remaining and not is_remaining_node_head and not is_new_head_provided:
        raise NoNewHeadError(f"Expected new head, but none was provided.")

    is_node_amount_one: bool = length(node=before.head) == 1
    is_nothing_between_anchors: bool = before.next is after and after.previous is before

    if is_node_amount_one or is_nothing_between_anchors:
        removed_sequence_start: None = None
        removed_sequence_end: None = None
    else:
        removed_sequence_start: Node = before.next
        removed_sequence_end: Node = after.previous

    is_new_sequence_provided: bool = insert is not None
    is_sequence_to_be_removed: bool = removed_sequence_start is not None and removed_sequence_end is not None

    if is_new_sequence_provided:
        before_next: Node = insert
        after_previous: Node = insert.previous
    else:
        before_next: Node = after
        after_previous: Node = before

    if is_new_sequence_provided and is_sequence_to_be_removed:
        stitch(head=before_next, last=before), \
            stitch(head=after, last=after_previous), \
            stitch(head=removed_sequence_start, last=removed_sequence_end)
    elif not is_new_sequence_provided and is_sequence_to_be_removed:
        stitch(head=before_next, last=before), \
            stitch(head=removed_sequence_start, last=removed_sequence_end)
    elif is_new_sequence_provided and not is_sequence_to_be_removed:
        stitch(head=before_next, last=before), \
            stitch(head=after, last=after_previous)
    elif not is_new_sequence_provided and not is_sequence_to_be_removed:
        pass
    else:
        raise ValueError(f"Unclear how Node sequences should be re-stitched...")

    if is_new_head_provided:
        update_head(node=head)
    elif is_new_sequence_provided or is_sequence_to_be_removed:
        update_from(node=before_next)

    if is_sequence_to_be_removed:
        update_head(node=removed_sequence_start)

    return removed_sequence_start


def update_from(node: Node) -> None:
    """
    Updates head and indices from node to and including end of sequence.
    Assumes that Node connections are consistent and only head and indices needs to be updated.
    """

    current: Node = node

    if length_step(current) == 1:
        head: Node = current
    else:
        head: Node = current.previous.head

    if current is head:
        index: int = 0
    else:
        index: int = current.previous.index + 1

    while True:
        current.head = head
        current.index = index

        current = current.next
        index += 1

        if current is head:
            break


def update_head(node: Node) -> None:
    """Update all Nodes with new head."""

    current: Node = node
    index: int = 0

    while True:
        current.head = node
        current.index = index

        current = current.next
        index += 1

        if current is node:
            break


def before_target(current: Node, head: Node, target: Node) -> Node:
    """
    When last node value is still smaller than target value, then the last node is "before target".
    Assuming pre-sorted sub-lists.
    Assuming first head value is lower than or equal to insert head value.
    """

    while current.next.value <= target.value and current.next is not head:
        current = current.next

    return current


def split_in_middle(head: Node) -> tuple[Node, Node]:
    before_head: Node = head
    _, after_head = middle_adjacent(head=head)

    split(first_head=before_head, second_head=after_head)

    return before_head, after_head


def merge(first: Node, second: Node) -> Node:
    """
    Merge pre-sorted circular doubly linked nodes.
    Moving nodes from second into sorted positions in first.
    Assuming pre-sorted sub-lists.
    """

    if first.value > second.value:
        first, second = second, first

    before_insert: Node = first

    while second is not None:
        insert, second = split_head_from_tail(node=second)
        before_insert = before_target(current=before_insert, head=first, target=insert)
        insert_between(before=before_insert, after=before_insert.next, insert=insert)

    return first


def merge_sort(head: Node) -> Node:
    """Merge-sort implementation for circular doubly linked nodes."""

    if head.next is head:
        # When there is only one value, there is nothing to sort.
        merged_sorted: Node = head
    else:
        first, second = split_in_middle(head=head)
        first_sorted, second_sorted = merge_sort(head=first), merge_sort(head=second)
        merged_sorted: Node = merge(first=first_sorted, second=second_sorted)

    return merged_sorted



import pytest

from src.py_cdll.exceptions import IndexNotFoundError, IndexOutOfRangeError
from src.py_cdll.node import Node, nodes_from_values, node_with_index, is_consistent, is_value_at_index


def test_node_with_index_one_node_index_zero_success():
    # Setup
    values0: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    node1: Node = node_with_index(node=node0, index=0)

    # Validation
    assert node0 is node1
    assert is_consistent(node=node0)
    assert is_value_at_index(node=node0, reference=values0)


def test_node_with_index_two_nodes_index_one_success():
    # Setup
    values0: list[str] = ["zero", "one"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    node1: Node = node_with_index(node=node0, index=1)

    # Validation
    assert is_consistent(node=node0)
    assert is_value_at_index(node=node0, reference=values0)
    assert is_value_at_index(node=node1, reference=values0)


def test_node_with_index_seven_nodes_index_five_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    node1: Node = node_with_index(node=node0, index=5)

    # Validation
    assert is_consistent(node=node0)
    assert is_value_at_index(node=node0, reference=values0)
    assert is_value_at_index(node=node1, reference=values0)


def test_node_with_index_seven_nodes_index_seven_out_of_range_failure():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six"]
    node0: Node = nodes_from_values(values=values0)

    # Validation
    with pytest.raises(IndexOutOfRangeError):
        _ = node_with_index(node=node0, index=7)


def test_node_with_index_four_nodes_index_minus_two_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    node0: Node = nodes_from_values(values=values0)

    # Validation
    with pytest.raises(IndexNotFoundError):
        _ = node_with_index(node=node0, index=-2)



from src.py_cdll import CDLL


def test__remove_with_slice_single_empty_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 3, 1)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_multiple_empty_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(1, 4, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_single_in_middle_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 3, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_multiple_in_middle_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(1, 4, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_single_from_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(4, 5, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_multiple_from_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(3, 5, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_multiple_spread_out_success():
    # Setup
    datas0: list[int] = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9]
    datas1: list[int] = [0, 1, 3, 4, 6, 7, 8, 9]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(2, 7, 3)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_all_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = []
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(0, None, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_nothing_after_last_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(5, None, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1


def test__remove_with_slice_nothing_after_last_with_high_index_success():
    # Setup
    datas0: list[int] = [1, 2, 3, 4, 5]
    datas1: list[int] = [1, 2, 3, 4, 5]
    cdll0: CDLL[int] = CDLL(values=datas0)
    cdll1: CDLL[int] = CDLL(values=datas1)
    slice0: slice = slice(555, None, None)

    # Execution
    cdll0._remove_with_slice(segment=slice0)

    # Validation
    assert cdll0 == cdll1



from src.py_cdll import CDLL
from src.py_cdll.node import Node, is_consistent, is_value_at_index


def test__node_replace_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "new_data"
    values0: list[str] = [data0, data1, data2]
    values1: list[str] = [data0, data3, data2]
    new_node: Node = Node(value=data3)
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0._node_replace(target=cdll0._head.next, replacement=new_node)

    # Validation
    assert data1 not in cdll0
    assert cdll0._head.next is new_node
    assert cdll0._head.next.next.value == data2
    assert cdll0._head.next.previous.value == data0
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test__node_replace_reassign_head_on_overwrite_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "new_data"
    values0: list[str] = [data0, data1, data2]
    values1: list[str] = [data3, data1, data2]
    new_node: Node = Node(value=data3)
    cdll0: CDLL = CDLL(values=values0)

    # Execution
    cdll0._node_replace(target=cdll0._head, replacement=new_node)

    # Validation
    assert data0 not in cdll0
    assert cdll0._head is new_node
    assert cdll0._head.next.next.value == data2
    assert cdll0._head.previous.value == data2
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)



import pytest

from src.py_cdll import CDLL
from src.py_cdll.node import Node, is_consistent, is_value_at_index, length


def test___setitem___index_looping_over_and_replacing_all_success():
    # Setup
    data0: int = 37
    data1: int = 73
    values0: list[int] = [data0, data1]
    values1: list[int] = [0, 1]
    cdll0: CDLL = CDLL(values=values0)
    node0: Node = cdll0._head
    node1: Node = cdll0._head.next

    # Verification
    assert node0.value == data0
    assert node1.value == data1
    assert is_consistent(cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values0)

    # Execution
    for index, _ in enumerate(cdll0):
        cdll0[index] = index

    # Validation
    assert cdll0._head.value == 0
    assert cdll0._last.value == 1
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last
    assert cdll0._head is cdll0._head.head
    assert is_consistent(cdll0._head)
    assert is_consistent(node0)
    assert is_consistent(node1)
    assert is_value_at_index(node=cdll0._head, reference=values1)
    assert is_value_at_index(node=node0, reference=values0[0:1])
    assert is_value_at_index(node=node1, reference=values0[1:])


def test___setitem___four_items_change_head_success():
    # Setup
    data: str = "data"
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    data3: str = "data3"
    values0: list[str] = [data0, data1, data2, data3]
    values1: list[str] = [data, data1, data2, data3]
    values2: list[str] = [data0]
    cdll0: CDLL = CDLL(values=values0)
    node0: Node = cdll0._head

    # Execution
    cdll0[0] = data

    # Validation
    assert cdll0._head.value is data
    assert cdll0._head is cdll0._head.head
    assert length(node=node0) == 1
    assert is_consistent(node0)
    assert is_consistent(cdll0._head)
    assert is_value_at_index(node=node0, reference=values2)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test___setitem___index_in_range_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])
    node0: Node = cdll0._head

    # Execution
    cdll0[0] = data1

    # Validation
    assert cdll0._head.value is data1
    assert cdll0._head is cdll0._head.head
    assert cdll0._head.next is cdll0._head
    assert cdll0._head.previous is cdll0._head
    assert is_consistent(cdll0._head)
    assert is_consistent(node0)


def test___setitem___index_out_of_range_failure():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    cdll0: CDLL = CDLL(values=[data0])

    # Validation
    with pytest.raises(IndexError):
        cdll0[1] = data1


def test___setitem___index_empty_list_out_of_range_failure():
    # Setup
    data0: str = "data0"
    cdll0: CDLL = CDLL()

    # Validation
    with pytest.raises(IndexError):
        cdll0[0] = data0


def test___setitem___index_negative_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1]
    cdll0: CDLL = CDLL(values=datas0)
    node0: Node = cdll0._head
    node1: Node = cdll0._head.next

    # Verification
    assert cdll0._head.value is data0
    assert cdll0._last.value is data1
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last
    assert cdll0._last.next is cdll0._head
    assert cdll0._last.previous is cdll0._head

    # Execution
    cdll0[-1] = data2

    # Validation
    assert cdll0._head.value is data0
    assert cdll0._last.value is data2
    assert cdll0._head is cdll0._head.head
    assert cdll0._head.next is cdll0._last
    assert cdll0._head.previous is cdll0._last
    assert cdll0._last.next is cdll0._head
    assert cdll0._last.previous is cdll0._head
    assert is_consistent(cdll0._head)
    assert is_consistent(node0)
    assert is_consistent(node1)



from re import match

from src.py_cdll import CDLL


def test___str___empty_success():
    # Setup
    cdll0: CDLL = CDLL()
    str0: str = r"^CDLL\[\]$"

    # Execution
    str1: str = cdll0.__str__()

    # Validation
    assert match(str0, str1) is not None


def test___str___text_success():
    # Setup
    data0: str = "data0"
    data1: str = "data1"
    data2: str = "data2"
    datas0: list[str] = [data0, data1, data2]
    cdll0: CDLL[str] = CDLL(values=datas0)
    str0: str = r"^CDLL\[" + f"'{data0}', '{data1}', '{data2}'" + r"\]$"

    # Execution
    str1: str = cdll0.__str__()

    # Validation
    assert match(str0, str1) is not None


def test___str___int_success():
    # Setup
    data0: int = 0
    data1: int = 1
    data2: int = 2
    datas0: list[int] = [data0, data1, data2]
    cdll0: CDLL[int] = CDLL(values=datas0)
    str0: str = r"^CDLL\[" + f"{data0}, {data1}, {data2}" + r"\]$"

    # Execution
    str1: str = cdll0.__str__()

    # Validation
    assert match(str0, str1) is not None


def test___str___float_success():
    # Setup
    data0: float = 0.1
    data1: float = 1.2
    data2: float = 2.3
    datas0: list[float] = [data0, data1, data2]
    cdll0: CDLL[float] = CDLL(values=datas0)
    str0: str = r"^CDLL\[" + f"{data0}, {data1}, {data2}" + r"\]$"

    # Execution
    str1: str = cdll0.__str__()

    # Validation
    assert match(str0, str1) is not None



from src.py_cdll.node import Node, nodes_from_values, is_value_at_index


def test_is_value_at_index_one_value_success():
    # Setup
    values0: list[str] = ["zero"]
    values1: list[str] = ["zero"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is True


def test_is_value_at_index_four_values_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three"]
    values1: list[str] = ["zero", "one", "two", "three"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is True


def test_is_value_at_index_seven_values_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four", "five", "six"]
    values1: list[str] = ["zero", "one", "two", "three", "four", "five", "six"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is True


def test_is_value_at_index_five_values_one_incorrect_value_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four"]
    values1: list[str] = ["zero", "one", "two", "incorrect", "four"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is False


def test_is_value_at_index_five_values_two_indices_mixed_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four"]
    values1: list[str] = ["zero", "one", "three", "two", "four"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is False


def test_is_value_at_index_five_values_three_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one", "two"]
    values1: list[str] = ["zero", "one", "two", "three", "four"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is False


def test_is_value_at_index_three_values_five_nodes_success():
    # Setup
    values0: list[str] = ["zero", "one", "two", "three", "four"]
    values1: list[str] = ["zero", "one", "two"]
    node0: Node = nodes_from_values(values=values0)

    # Execution
    bool0: bool = is_value_at_index(node=node0, reference=values1)

    # Validation
    assert bool0 is False



import pytest

from src.py_cdll import NoNewHeadError
from src.py_cdll.node import Node, insert_between, nodes_from_values, equal, is_consistent, is_value_at_index, length


def test_insert_between_first_zero_second_one_success():
    # Setup
    values0: list[int] = [0]
    values1: list[int] = [1]
    values2: list[int] = [0, 1]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0
    after0: Node = head0.next
    insert0: Node = head1
    removed0: None = None

    # Execution
    removed1: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed1 is removed0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_seven_second_none_success():
    # Setup
    values0: list[int] = [7]
    head0: Node = nodes_from_values(values=values0)
    head1: None = None
    head2: Node = nodes_from_values(values=values0)
    before0: Node = head0
    after0: Node = head0.next
    insert0: None = head1

    # Execution
    inserted0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert inserted0 is insert0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_value_at_index(node=head0, reference=values0) is True
    assert is_value_at_index(node=head2, reference=values0) is True


def test_insert_between_first_zero_two_second_one_success():
    # Setup
    values0: list[int] = [0, 2]
    values1: list[int] = [1]
    values2: list[int] = [0, 1, 2]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0
    after0: Node = head0.next
    insert0: Node = head1
    removed0: None = None

    # Execution
    removed1: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed1 is removed0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_five_seven_second_three_replace_head_success():
    # Setup
    values0: list[int] = [5, 7]
    values1: list[int] = [3]
    values2: list[int] = [3, 7]
    values3: list[int] = [5]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0.previous
    after0: Node = head0.next
    insert0: Node = head1
    removed0: Node = head0

    # Execution
    removed1: Node = insert_between(before=before0, after=after0, insert=insert0, head=head1)

    # Validation
    assert removed1 is removed0
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_value_at_index(node=head0, reference=values3) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_five_seven_nine_second_three_replace_head_success():
    # Setup
    values0: list[int] = [5, 7, 9]
    values1: list[int] = [3]
    values2: list[int] = [3, 7, 9]
    values3: list[int] = [5]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0.previous
    after0: Node = head0.next
    insert0: Node = head1
    removed0: Node = head0

    # Execution
    removed1: Node = insert_between(before=before0, after=after0, insert=insert0, head=head1)

    # Validation
    assert removed1 is removed0
    assert length(node=head0) == 1
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_consistent(node=removed1) is True
    assert is_value_at_index(node=head0, reference=values3) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_five_seven_nine_ten_second_three_replace_head_success():
    # Setup
    values0: list[int] = [5, 7, 9, 10]
    values1: list[int] = [3]
    values2: list[int] = [3, 7, 9, 10]
    values3: list[int] = [5]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0.previous
    after0: Node = head0.next
    insert0: Node = head1
    removed0: Node = head0

    # Execution
    removed1: Node = insert_between(before=before0, after=after0, insert=insert0, head=head1)

    # Validation
    assert removed1 is removed0
    assert length(node=head0) == 1
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_consistent(node=removed1) is True
    assert is_value_at_index(node=head0, reference=values3) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_three_nine_second_none_success():
    # Setup
    values0: list[int] = [3, 9]
    head0: Node = nodes_from_values(values=values0)
    head1: None = None
    head2: Node = nodes_from_values(values=values0)
    before0: Node = head0
    after0: Node = head0.next
    insert0: None = head1

    # Execution
    inserted0: None = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert inserted0 is insert0
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_value_at_index(node=head0, reference=values0) is True
    assert is_value_at_index(node=head2, reference=values0) is True


def test_insert_between_first_three_nine_second_none_remove_success():
    # Setup
    values0: list[int] = [3, 9]
    values1: list[int] = [9]
    values2: list[int] = [3]
    head0: Node = nodes_from_values(values=values0)
    head1: None = None
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0
    after0: Node = head0
    insert0: None = head1
    removed0: Node = head0.next

    # Execution
    removed1: Node = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert removed1 is removed0
    assert equal(first=removed1, second=removed0)
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=removed1) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True
    assert is_value_at_index(node=removed1, reference=values1) is True


def test_insert_between_first_five_twelve_second_two_success():
    # Setup
    values0: list[int] = [5, 12]
    values1: list[int] = [2]
    values2: list[int] = [2, 5, 12]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0.previous
    after0: Node = head0
    insert0: Node = head1
    removed0: None = None

    # Execution
    removed1: None = insert_between(before=before0, after=after0, insert=insert0, head=insert0)

    # Validation
    assert removed1 is removed0
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_five_twelve_second_two_four_success():
    # Setup
    values0: list[int] = [5, 12]
    values1: list[int] = [2, 4]
    values2: list[int] = [2, 4, 5, 12]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0.previous
    after0: Node = head0
    insert0: Node = head1
    removed0: None = None

    # Execution
    removed1: None = insert_between(before=before0, after=after0, insert=insert0, head=insert0)

    # Validation
    assert removed1 is removed0
    assert equal(first=head1, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True


def test_insert_between_first_five_twelve_sixteen_twenty_second_none_success():
    # Setup
    values0: list[int] = [5, 12, 16, 20]
    values1: list[int] = [12, 16]
    values2: list[int] = [5, 20]
    head0: Node = nodes_from_values(values=values0)
    head1: None = None
    head2: Node = nodes_from_values(values=values2)
    before0: Node = head0
    after0: Node = head0.previous
    insert0: None = head1
    removed0: Node = nodes_from_values(values=values1)

    # Execution
    inserted0: Node = insert_between(before=before0, after=after0, insert=insert0)

    # Validation
    assert equal(first=inserted0, second=removed0)
    assert equal(first=head0, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=inserted0) is True
    assert is_value_at_index(node=head0, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True
    assert is_value_at_index(node=removed0, reference=values1) is True


def test_insert_between_first_three_seven_nine_fifteen_nineteen_second_ten_twelve_two_five_without_head_failure():
    # Setup
    head0: Node = nodes_from_values(values=[3, 7, 9, 15, 19])
    head1: Node = nodes_from_values(values=[10, 12, 2, 5])
    before0: Node = head0.next.next
    after0: Node = before0
    insert0: Node = head1

    # Validation
    with pytest.raises(NoNewHeadError):
        _: Node = insert_between(before=before0, after=after0, insert=insert0)


def test_insert_between_first_three_seven_nine_fifteen_nineteen_second_ten_twelve_two_five_with_head_success():
    # Setup
    values0: list[int] = [3, 7, 9, 15, 19]
    values1: list[int] = [10, 12, 2, 5]
    values2: list[int] = [2, 5, 9, 10, 12]
    values3: list[int] = [15, 19, 3, 7]
    head0: Node = nodes_from_values(values=values0)
    head1: Node = nodes_from_values(values=values1)
    head2: Node = nodes_from_values(values=values2)
    head3: Node = head1.previous.previous
    before0: Node = head0.next.next
    after0: Node = before0
    insert0: Node = head1
    removed1: Node = nodes_from_values(values=values3)

    # Execution
    inserted0: Node = insert_between(before=before0, after=after0, insert=insert0, head=head3)

    # Validation
    assert equal(first=inserted0, second=removed1)
    assert equal(first=head3, second=head2)
    assert is_consistent(node=head0) is True
    assert is_consistent(node=head1) is True
    assert is_consistent(node=inserted0) is True
    assert is_value_at_index(node=head0, reference=values3) is True
    assert is_value_at_index(node=head1, reference=values2) is True
    assert is_value_at_index(node=head2, reference=values2) is True
    assert is_value_at_index(node=removed1, reference=values3) is True



from src.py_cdll import CDLL
from src.py_cdll.node import Node, is_consistent, is_value_at_index


def test__node_remove_from_single_success():
    # Setup
    values0: list[int] = [0]
    cdll0: CDLL = CDLL(values=values0)
    node0: Node = cdll0._head

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert node0 not in cdll0._nodes()
    assert len(cdll0) == 0
    assert cdll0.is_empty() is True


def test__node_remove_head_from_triple_success():
    # Bug: _node_remove had no check for whether head was being removed, which would invalidate pointer in CDLL
    # and cause infinite loops when trying to find head after that.

    # Setup
    values0: list[int] = [0, 1, 2]
    values1: list[int] = [1, 2]
    cdll0: CDLL = CDLL(values=values0)
    node0: Node = cdll0._head

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert node0 not in cdll0._nodes()
    assert len(cdll0) == 2
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)


def test__node_remove_from_multiple_success():
    # Setup
    values0: list[int] = [0, 1, 2, 3, 4]
    values1: list[int] = [0, 1, 2, 4]
    cdll0: CDLL = CDLL(values=values0)
    node0: Node = cdll0._head.previous.previous

    # Execution
    cdll0._node_remove(target=node0)

    # Validation
    assert node0 not in cdll0._nodes()
    assert len(cdll0) == 4
    assert cdll0._head is cdll0._head.head
    assert is_consistent(node=cdll0._head)
    assert is_value_at_index(node=cdll0._head, reference=values1)
